from itertools import combinations
import time
import pandas as pd
import numpy as np

def correct_magnitudes(df, correction_pairs):    
    '''
    Correct the magnitudes of a set of filters inside a dataframe
    
    Keyword Arguments:
    df - Dataframe with uncorrected magnitudes
    correction_pairs - Dictionary with all the combinations of filters and corrections to be applied
    '''
    
    for filt in correction_pairs:
        df[filt] = df[filt] - df[correction_pairs[filt]]
    
    return df


def create_colors(df, filters):
    '''
    Create all the possible filter combinations (colors) for a set of filters inside a dataframe
    
    Keyword arguments:
    df - Dataframe with the magnitudes
    filters - List of filters to combine
    '''
    
    comb_list = list(combinations(filters, 2))
    colors_df = pd.DataFrame()
    
    for comb in comb_list:
        color_name = f'({comb[0]} - {comb[1]})'
        color_values = (df[comb[0]] - df[comb[1]])
        
        colors_df = pd.concat([colors_df, color_values.rename(color_name)], axis = 1)

    return colors_df
        

def create_combinations(df, filters):
    '''
    Create all the possible color combinations for a set of filters inside a dataframe
    
    Keyword arguments:
    df - Dataframe with the magnitudes
    filters - Set of filters to combine
    '''
    
    comb_list = list(combinations(filters, 4))
    combinations_df = pd.DataFrame()
    
    for comb in comb_list:
        combination_name = f'({comb[0]} - {comb[1]}) - ({comb[2]} - {comb[3]})'
        combination_value = (df[comb[0]] - df[comb[1]]) - (df[comb[2]] - df[comb[3]])
        
        combinations_df = pd.concat([combinations_df, combination_value.rename(combination_name)], axis = 1)

    return combinations_df


def assemble_work_df(df, filters, correction_pairs, add_colors = False, add_combinations = False):
    '''
    Assemble a dataframe with a set of magnitudes and, when asked, colors and color combinations.
    
    Keyword arguments:
    df - Dataframe with the magnitudes
    
    filters - Set of filters to use
    correction_pairs - Dictionary with the pairs of filters and corrections to be applied
    
    add_colors - If True, all the possible colors will be added to the returned dataframe
    add_combinations - If True all the possible color combinations will be added to the returned dataframe
    '''

    print('Iniciando processo de criação do dataframe de trabalho:\n')
    
    # If a dictionary of corrections is passed, apply the corrections to the magnitudes
    if correction_pairs:
        print('  - Aplicando correção de magnitudes...', end = '')
        
        start_time = time.time()
        df = correct_magnitudes(df, correction_pairs)
        print(f' Tempo: {(time.time() - start_time):.2f} s')
    
    # Filter the df to get only the filters passed
    work_df = df[filters].copy()
    
    #If asked for, create a dataframe with all the possible colors and add it to the work dataframe
    if add_colors is True:
        print('  - Adicionando cores ao dataframe...', end = '')
        
        start_time = time.time()
        colors_df = create_colors(work_df, filters)
        work_df = pd.concat([work_df, colors_df], axis = 1)
        print(f' Tempo: {(time.time() - start_time):.2f} s')
        
    #If asked for, create a dataframe with all the possible color combinations and add it to the work dataframe
    if add_combinations is True:
        print('  - Adicionando combinações de cores ao dataframe...', end = '')
        
        start_time = time.time()
        combinations_df = create_combinations(work_df, filters)
        work_df = pd.concat([work_df, combinations_df], axis = 1)
        print(f' Tempo: {(time.time() - start_time):.2f} s')
    
    print(f'\nProcesso finalizado! Shape da Tabela Gerada: {work_df.shape}')
    
    #Return the resulting dataframe
    return work_df


def calculate_abs_mag(df, filters, distance):
    '''
    Calculate the absolute magnitudes of a set of apparent magnitudes inside a dataframe.
    
    Keyword arguments:
    df - Dataframe with the apparent magnitudes and distances
    
    filters - Set of filters to use
    distance - Name of the distance column to use
    '''
    
    y = 5 * (df[distance].apply(np.log10) - 1)
    
    for filt in filters:
        df[filt] = df[filt] - y

    return df


from ._functions import (correct_magnitudes,
                         create_colors,
                         create_combinations,
                         assemble_work_df,
                         calculate_abs_mag)


import numpy as np
import pandas as pd

def calculate_mad(predictions, true_values, bins, param_unit):
    
    if not bins:
        bins = [min(true_values) - 1, max(true_values) + 1]
    
    df = pd.merge(left = true_values, left_index = True,
                  right = predictions, right_index = True)

    df.columns = ['TRUE_VALUE', 'PREDICTION']

    bins_intervals = []
    bins_sizes = []
    bins_mads = []

    for bin_index in range(0, len(bins) - 1):
        bin_min = bins[bin_index]
        bin_max = bins[bin_index + 1]
        bins_intervals.append(f'[{bin_min} {param_unit}, {bin_max} {param_unit}]')

        df_bin = df[(df['PREDICTION'] >= bin_min) & (df['PREDICTION'] < bin_max)].copy()
        bins_sizes.append(df_bin.shape[0])

        errors = df_bin['PREDICTION'] - df_bin['TRUE_VALUE']
        mad = np.median(np.abs(errors))
        bins_mads.append(mad)


    bins_intervals.append('Full Sample')
    bins_sizes.append(df.shape[0])
    error = df['PREDICTION'] - df['TRUE_VALUE']
    mad = np.median(np.abs(error))
    bins_mads.append(mad)

    return pd.DataFrame({'bin': bins_intervals,
                         'objects': bins_sizes,
                         'mad': bins_mads})


import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
import pandas as pd
import numpy as np

from math import log10, floor

label_fontdict = {'family': 'sans-serif',
                  'weight': 'normal',
                  'size': 16}

legend_label_fontdict = {'family': 'sans-serif',
                         'weight': 'normal',
                         'size': 14}

legend_title_fontdict = {'family': 'sans-serif',
                         'weight': 'bold',
                         'size': 18}

def plot_test_graphs(predictions, true_values, bins, cmap_name, param_string, param_unit, n_ticks, legend):

    if not bins:
        bins = [min(true_values) - 1, max(true_values) + 1]

    df = pd.merge(left = true_values, left_index = True,
                  right = predictions, right_index = True)

    df.columns = ['TRUE_VALUE', 'PREDICTION']
    df['ERROR'] = df['PREDICTION'] - df['TRUE_VALUE']

    fig, ax = plt.subplots(nrows = 1, ncols = 2,
                           width_ratios = [0.6, 0.4],
                           figsize = (15, 6))

    colors = mpl.colormaps[cmap_name](np.linspace(0.5, 1.00, len(bins) - 1))
    
    bins_intervals = []
    for bin_index in range(0, len(bins) - 1):
        bin_min = bins[bin_index]
        bin_max = bins[bin_index + 1]
        bins_intervals.append(f'[{bin_min} {param_unit}, {bin_max} {param_unit}]')

        df_bin = df[(df['PREDICTION'] >= bin_min) & (df['PREDICTION'] < bin_max)].copy()

        sns.scatterplot(data = df_bin, x = 'PREDICTION', y = 'TRUE_VALUE',
                        ax = ax[0], s = 9, color = colors[bin_index],
                        linewidth = 0, zorder = 2)
        sns.kdeplot(data = df_bin, x = 'ERROR', ax = ax[1],
                    color = colors[bin_index])
        
    f = lambda m,c: ax[0].plot([],[],marker=m, color=c, ls="none")[0]
    handles = [f("s", colors[i]) for i in range(len(bins_intervals))]

    min_lim = round_to_1(bins[0] - (bins[-1] - bins[0]) * 0.05)
    max_lim = round_to_1(bins[-1] + (bins[-1] - bins[0]) * 0.05)

    ax[0].plot([min_lim, max_lim],
               [min_lim, max_lim],
               ls = '--',
               lw = 1.5,
               color = 'k',
               zorder = 3)

    ax[0] = beautify_graph(ax = ax[0],
                           x_limits = [min_lim, max_lim],
                           y_limits = [min_lim, max_lim],
                           n_ticks = n_ticks,
                           x_label = f'Predicted {param_string}',
                           y_label = f'True {param_string}')


    min_lim = round_to_1(-(df['ERROR'].abs().median() * 20))
    max_lim = round_to_1((df['ERROR'].abs().median() * 20))

    ax[1] = beautify_graph(ax = ax[1],
                           x_limits = [min_lim, max_lim],
                           y_limits = None,
                           n_ticks = n_ticks,
                           x_label = f'Error',
                           y_label = f'Density')

    ax[1].tick_params(axis = 'y', labelsize = 0)
    
    if legend:
        leg = fig.legend(handles,
                        bins_intervals,
                        title = f'{param_string}',
                        title_fontproperties = legend_title_fontdict,
                        ncols = len(bins_intervals),
                        loc = 'upper center',
                        bbox_to_anchor = (0.5, -0.075),
                        framealpha = 1,
                        prop = legend_label_fontdict,
                        markerscale = 3,
                        borderpad = 1)
        
        leg._legend_box.sep = 20
    
    return fig


def beautify_graph(ax, x_limits, y_limits, n_ticks, x_label, y_label):
    if x_limits:
        ax.set_xlim(x_limits[0], x_limits[1])
        ax.set_xticks(np.linspace(x_limits[0],
                                  x_limits[1],
                                  n_ticks))
        
    if y_limits:
        ax.set_ylim(y_limits[0], y_limits[1])
        ax.set_yticks(np.linspace(y_limits[0],
                                  y_limits[1],
                                  n_ticks))

    ax.tick_params(labelsize = 14)

    ax.set_xlabel(x_label,
                  fontdict = label_fontdict,
                  labelpad = 15)
    ax.set_ylabel(y_label,
                  fontdict = label_fontdict,
                  labelpad = 15)

    ax.grid(zorder = 0)
    
    return ax


def round_to_1(x):
    return round(x, -int(floor(log10(abs(x)))))